#![allow(missing_copy_implementations)]

use core::ffi::c_void;
use azul_css::{AzString, StringVec, ColorU};
use azul_core::window::AzStringPair;
use tinyfiledialogs::{MessageBoxIcon, DefaultColorValue};

/// Ok or cancel result, returned from the `msg_box_ok_cancel` function
#[derive(Debug)]
pub struct MsgBox {
    /// reserved pointer (currently nullptr) for potential C extension
    pub _reserved: *mut c_void,
}

/// Ok or cancel result, returned from the `msg_box_ok_cancel` function
#[derive(Debug)]
pub struct FileDialog {
    /// reserved pointer (currently nullptr) for potential C extension
    pub _reserved: *mut c_void,
}

/// Ok or cancel result, returned from the `msg_box_ok_cancel` function
#[derive(Debug)]
pub struct ColorPickerDialog {
    /// reserved pointer (currently nullptr) for potential C extension
    pub _reserved: *mut c_void,
}


#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
#[repr(C)]
pub enum OkCancel {
    Ok,
    Cancel,
}

impl From<::tinyfiledialogs::OkCancel> for OkCancel {
    #[inline]
    fn from(e: ::tinyfiledialogs::OkCancel) -> OkCancel {
        match e {
            ::tinyfiledialogs::OkCancel::Ok => OkCancel::Ok,
            ::tinyfiledialogs::OkCancel::Cancel => OkCancel::Cancel,
        }
    }
}

impl From<OkCancel> for ::tinyfiledialogs::OkCancel {
    #[inline]
    fn from(e: OkCancel) -> ::tinyfiledialogs::OkCancel {
        match e {
            OkCancel::Ok => ::tinyfiledialogs::OkCancel::Ok,
            OkCancel::Cancel => ::tinyfiledialogs::OkCancel::Cancel,
        }
    }
}

/// "Ok / Cancel" MsgBox (title, message, icon, default)
pub fn msg_box_ok_cancel(title: &str, message: &str, icon: MessageBoxIcon, default: OkCancel) -> OkCancel {
    ::tinyfiledialogs::message_box_ok_cancel(title, message, icon, default.into()).into()
}

/// Yes or No result, returned from the `msg_box_yes_no` function
#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
#[repr(C)]
pub enum YesNo {
    Yes,
    No,
}

impl From<YesNo> for ::tinyfiledialogs::YesNo {
    #[inline]
    fn from(e: YesNo) -> ::tinyfiledialogs::YesNo {
        match e {
            YesNo::Yes => ::tinyfiledialogs::YesNo::Yes,
            YesNo::No => ::tinyfiledialogs::YesNo::No,
        }
    }
}

impl From<::tinyfiledialogs::YesNo> for YesNo {
    #[inline]
    fn from(e: ::tinyfiledialogs::YesNo) -> YesNo {
        match e {
            ::tinyfiledialogs::YesNo::Yes => YesNo::Yes,
            ::tinyfiledialogs::YesNo::No => YesNo::No,
        }
    }
}

/// MsgBox icon to use in the `msg_box_*` functions
#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
#[repr(C)]
pub enum MsgBoxIcon {
    Info,
    Warning,
    Error,
    Question,
}

impl From<MsgBoxIcon> for MessageBoxIcon {
    #[inline]
    fn from(e: MsgBoxIcon) -> MessageBoxIcon {
        match e {
            MsgBoxIcon::Info => MessageBoxIcon::Info,
            MsgBoxIcon::Warning => MessageBoxIcon::Warning,
            MsgBoxIcon::Error => MessageBoxIcon::Error,
            MsgBoxIcon::Question => MessageBoxIcon::Question,
        }
    }
}

// Note: password_box, input_box and list_dialog do not work, so they're not included here.

/// "Y/N" MsgBox (title, message, icon, default)
pub fn msg_box_yes_no(title: &str, message: &str, icon: MessageBoxIcon, default: YesNo) -> YesNo {
    ::tinyfiledialogs::message_box_yes_no(title, message, icon, default.into()).into()
}

/// "Ok" MsgBox (title, message, icon)
pub fn msg_box_ok(title: &str, message: &str, icon: MessageBoxIcon) {
    ::tinyfiledialogs::message_box_ok(title, message, icon)
}

/// Wrapper around `message_box_ok` with the default title "Info" + an info icon.
pub fn msg_box(content: &str) {
    msg_box_ok("Info", content, MessageBoxIcon::Info);
}

/// Opens the default color picker dialog
pub fn color_picker_dialog(title: &str, default_value: Option<ColorU>) -> Option<ColorU> {
    let rgb = [
        default_value.map(|c| c.r).unwrap_or_default(),
        default_value.map(|c| c.g).unwrap_or_default(),
        default_value.map(|c| c.b).unwrap_or_default(),
    ];

    let default = DefaultColorValue::RGB(&rgb);
    let result = ::tinyfiledialogs::color_chooser_dialog(title, default)?;
    Some(ColorU { r: result.1[0], g: result.1[1], b: result.1[2], a: ColorU::ALPHA_OPAQUE })
}

#[derive(Debug, Clone, PartialEq, PartialOrd)]
#[repr(C)]
pub struct FileTypeList {
    pub document_types: StringVec,
    pub document_descriptor: AzString,
}

impl_option!(FileTypeList, OptionFileTypeList, copy = false, [Debug, Clone, PartialEq, PartialOrd]);

/// Open a single file, returns `None` if the user canceled the dialog.
///
/// Filters are the file extensions, i.e. `Some(&["doc", "docx"])` to only allow
/// "doc" and "docx" files
pub fn open_file_dialog(title: &str, default_path: Option<&str>, filter_list: Option<FileTypeList>) -> Option<AzString> {
    let documents: Vec<AzString> = filter_list.as_ref().map(|s| s.document_types.clone().into_library_owned_vec()).unwrap_or_default().into();
    let documents: Vec<&str> = documents.iter().map(|s| s.as_str()).collect();
    let filter_list_ref = match filter_list.as_ref() {
        Some(s) => Some((documents.as_ref(), s.document_descriptor.as_str())),
        None => None,
    };
    let path = default_path.unwrap_or("");
    ::tinyfiledialogs::open_file_dialog(title, path, filter_list_ref).map(|s| s.into())
}

/// Open a directory, returns `None` if the user canceled the dialog
pub fn open_directory_dialog(title: &str, default_path: Option<&str>)
-> Option<AzString>
{
    ::tinyfiledialogs::select_folder_dialog(title, default_path.unwrap_or("")).map(|s| s.into())
}

/// Open multiple files at once, returns `None` if the user canceled the dialog,
/// otherwise returns the `Vec<String>` with the given file paths
///
/// Filters are the file extensions, i.e. `Some(&["doc", "docx"])` to only allow
/// "doc" and "docx" files
pub fn open_multiple_files_dialog(title: &str, default_path: Option<&str>, filter_list: Option<FileTypeList>)
-> Option<StringVec>
{
    let documents: Vec<AzString> = filter_list.as_ref().map(|s| s.document_types.clone().into_library_owned_vec()).unwrap_or_default().into();
    let documents: Vec<&str> = documents.iter().map(|s| s.as_str()).collect();
    let filter_list_ref = match filter_list.as_ref() {
        Some(s) => Some((documents.as_ref(), s.document_descriptor.as_str())),
        None => None,
    };
    let path = default_path.unwrap_or("");
    ::tinyfiledialogs::open_file_dialog_multi(title, path, filter_list_ref).map(|s| s.into())
}

/// Opens a save file dialog, returns `None` if the user canceled the dialog
pub fn save_file_dialog(title: &str, default_path: Option<&str>)
-> Option<AzString>
{
    let path = default_path.unwrap_or("");
    ::tinyfiledialogs::save_file_dialog(title, path).map(|s| s.into())
}

// TODO (at least on Windows):
// - Find and replace dialog
// - Font picker dialog
// - Page setup dialog
// - Print dialog
// - Print property dialog
